from marnadi import Header, descriptors
from marnadi.utils import to_bytes


class HttpError(Exception):

    headers = descriptors.Headers(
        ('Content-Type', Header('text/plain', charset='utf-8')),
    )

    def __init__(self, status='500 Internal Server Error',
                 data=None, headers=None):
        self.status = status
        self.data = data
        if headers:
            self.headers.extend(headers)

    def __len__(self):
        return 1

    def __iter__(self):
        yield to_bytes(self.data)



import logging
import types

from marnadi import descriptors, Header
from marnadi.errors import HttpError
from marnadi.utils import metaclass, to_bytes


class Handler(type):

    __func__ = None

    def __call__(cls, *args, **kwargs):
        func = cls.__func__
        if func is not None:
            return func(*args, **kwargs)
        return super(Handler, cls).__call__(*args, **kwargs)

    def get_instance(cls, *args, **kwargs):
        return type.__call__(cls, *args, **kwargs)

    def provider(cls, func):
        assert callable(func)
        attributes = dict(
            __func__=staticmethod(func),
            __module__=func.__module__,
            __doc__=func.__doc__,
        )
        return type(cls)(func.__name__, (cls, ), attributes)

    def start(cls, **kwargs):
        raise NotImplementedError


@metaclass(Handler)
class Response(object):

    __slots__ = 'request', '__weakref__'

    logger = logging.getLogger('marnadi')

    supported_http_methods = {
        'OPTIONS', 'GET', 'HEAD', 'POST', 'PUT', 'PATCH', 'DELETE',
    }

    status = '200 OK'

    headers = descriptors.Headers(
        ('Content-Type', Header('text/plain', charset='utf-8')),
    )

    cookies = descriptors.Cookies()

    def __init__(self, request):
        self.request = request

    def __call__(self, **kwargs):
        if self.request.method not in self.supported_http_methods:
            raise HttpError(
                '501 Not Implemented',
                headers=(('Allow', ', '.join(self.allowed_http_methods)), )
            )
        callback = getattr(self, self.request.method.lower()) or self.__func__
        if callback is None:
            raise HttpError(
                '405 Method Not Allowed',
                headers=(('Allow', ', '.join(self.allowed_http_methods)), )
            )
        return callback(**kwargs)

    @classmethod
    def start(cls, **kwargs):
        """Start response with given params.

        Note:
            Error responses can be customized by overriding this method.
            For example your version may catch `HttpError` from original
            implementation and reraise it with necessary content data
            (which may be a HTML containing formatted stack trace).
        """
        request, start_response = yield
        try:
            response = cls.get_instance(request)
            result = response(**kwargs)
            if isinstance(result, types.GeneratorType):
                body = (
                    to_bytes(chunk, error_callback=cls.logger.exception)
                    for chunk in result
                )
            else:
                body = (to_bytes(result), )
                response.headers.setdefault('Content-Length', len(body[0]))
            start_response(
                response.status,
                list(response.headers.items(stringify=True)),
            )
            yield body
        except HttpError:
            raise
        except Exception as error:
            cls.logger.exception(error)
            raise HttpError

    @property
    def allowed_http_methods(self):
        func = self.__func__
        for method in self.supported_http_methods:
            if func or getattr(self, method.lower()):
                yield method

    def options(self, **kwargs):
        self.headers['Allow'] = ', '.join(self.allowed_http_methods)

    get = None

    head = None

    post = None

    put = None

    patch = None

    delete = None



import re

from marnadi.utils import Lazy


class Route(object):

    __slots__ = 'path', 'handler', 'params', 'pattern', 'name'

    placeholder_re = re.compile(r'\{([a-zA-Z_][a-zA-Z0-9_]*)\}')

    def __init__(self, path, handler, name=None, params=None, patterns=None):
        self.path = path
        self.handler = Lazy(handler)
        self.name = name
        self.params = params or {}
        self.pattern = self.make_pattern(path, patterns)

    def match(self, request_path):
        if self.pattern is not None:
            match = self.pattern.match(request_path)
            if match:
                return request_path[match.end(0):], match.groupdict()
        elif request_path.startswith(self.path):
            return request_path[len(self.path):], ()

    @classmethod
    def make_pattern(cls, path, placeholder_patterns=None):
        unescaped_path = path.replace('{{', '').replace('}}', '')
        placeholders = cls.placeholder_re.findall(unescaped_path)
        if not placeholders:
            return
        placeholder_patterns = placeholder_patterns or {}
        pattern = re.escape(path.replace('{{', '{').replace('}}', '}'))
        for placeholder in placeholders:
            pattern = pattern.replace(
                r'\{{{placeholder}\}}'.format(placeholder=placeholder),
                r'(?P<{name}>{pattern})'.format(
                    name=placeholder,
                    pattern=placeholder_patterns.get(placeholder, r'\w+')
                ),
            )
        return re.compile(pattern)

    def restore_path(self, **params):
        return self.path.format(**params)


class Header(object):

    __slots__ = 'value', 'params'

    def __init__(self, _value, **params):
        self.value = _value
        self.params = params

    def __str__(self):
        return self.make_value()

    def __bytes__(self):
        value = self.make_value()
        if isinstance(value, bytes):  # python 2.x
            return value
        return value.encode(encoding='latin1')

    def make_value(self):
        if not self.params:
            return str(self.value)
        return '{value}; {params}'.format(
            value=self.value,
            params='; '.join(
                '%s=%s' % (attr_name, attr_value)
                for attr_name, attr_value in self.params.items()
            ),
        )
